/*
  This is a simple Xwindow that displays live video

  This one tries to use a lot of fast techniques
*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <math.h>
#include <string.h>
#include <visenv.h>
#include <visX.h>

/*
  Global Variables determining window and display
*/
Display *curDisplay;
Window   curWindow;
GC       curGC; /* graphics context */
XImage  *xim1, *xim2;
int     defdepth,bitmap_pad;
int     bytesPerPixel;
Pixmap  gblPixmap;
Colormap curMap;

int Xsize=288;
int Ysize=216;

/*
  Global Variables for userland stuff
*/

unsigned char *ximdata;

// just...easier this way.

char list[MAXIMS][MAXBUF];
char source[MAXBUF];
Pixel thresh;
int threshpad;
int grass;
Image *theImage;

//prototypes
int initWindow(long nFrames);
void refreshWindow(int i);
void closeWindow(long nFrames);
/**********************************
 *       XLIB STUFF
 **********************************/

int initWindow(long nFrames) {
   int defScreen,i;
   XSetWindowAttributes wAttr;
   XGCValues gcValues;

   char buffer[64] = "Training Data Processor";
   XTextProperty xtp = {buffer, 125, 8, strlen(buffer)};


   /*
    * connect to the X server.  uses the server specified in the
    * DISPLAY environment variable
    */
   curDisplay = XOpenDisplay((char *) NULL);
   if ((Display *) NULL == curDisplay) {
      fprintf(stderr, "XRobo-View:  could not open display.\n");
      exit(-1);
   }
   
   /*
    * begin to create a window
    */
   printf("creating window\n");
   defdepth = DefaultDepth(curDisplay,0);
   bytesPerPixel = defdepth/8;
   bytesPerPixel = bytesPerPixel == 3 ? 4 : bytesPerPixel;
   printf("default depth %d\n", defdepth);

   defScreen = DefaultScreen(curDisplay);
   printf("Default screen %d\n", defScreen);

   curWindow = XCreateWindow(curDisplay, DefaultRootWindow(curDisplay),
			     10, 10, Xsize, Ysize, 0, 
			     defdepth, InputOutput, 
			     DefaultVisual(curDisplay, defScreen),
			     0, &wAttr);

   printf("Using depth %d\n", DefaultDepth(curDisplay, 0));

   /*
    * request mouse button and keypress events
    */
   wAttr.event_mask = ButtonPressMask | KeyPressMask | ExposureMask | ResizeRedirectMask;
   XChangeWindowAttributes(curDisplay, curWindow, CWEventMask, &wAttr);

   /*
    * force it to appear on the screen
    */
   XSetWMName(curDisplay, curWindow, &xtp);
   XStoreName(curDisplay,curWindow,"Neural Network Training Data Processor\0");
   XMapWindow(curDisplay, curWindow);

   /*
    * create a graphics context.  this stores a drawing state; stuff like
    * current color and line width.  this gc is for drawing into our
    * window. 
    */
   printf("creating graphics context\n");
   curGC = XCreateGC(curDisplay, curWindow, 0, &gcValues);

   XSetWindowColormap(curDisplay,curWindow,DefaultColormapOfScreen(DefaultScreenOfDisplay(curDisplay)));
   curMap=DefaultColormapOfScreen(DefaultScreenOfDisplay(curDisplay));
   //   curScreen=DefaultScreen(curDisplay);
   /*
    * create an XImage for holding image data
    */

   ximdata = malloc(Xsize * Ysize * bytesPerPixel);
   for(i=0;i<Xsize*Ysize*bytesPerPixel;i++)
     ximdata[i] = 0x00;
   
   bitmap_pad = defdepth == 24 ? 32 : defdepth;
   
   printf("creating frame 1 with depth %d\n", defdepth);
   xim1 = XCreateImage(curDisplay, DefaultVisual(curDisplay, defScreen), defdepth, 
		       ZPixmap, 0, ximdata, Xsize, Ysize, bitmap_pad, 0);
   
   printf("exiting init window\n");
   return(bytesPerPixel);
}

void refreshWindow(int i){
  char buffer[MAXBUF];
  int temp=255;
  Image * tempimg;

  sprintf(buffer,"%s/%s.ppm",source,list[i]);
  theImage->image=readPPM(&theImage->rows,&theImage->cols,&temp,buffer);
  theImage->imagesize=theImage->rows*theImage->cols;
  
  /* create the training image*/
  gaussianIm(theImage);
  cropXYIm(theImage, 16, 12);
  
  tempimg=theImage;
  theImage=chromaticityIm(tempimg);
  free(tempimg->image);
  free(tempimg);
  threshByPaddedColor(theImage,thresh, threshpad);
  if(grass>0){
    grassFire(theImage, (Pixel){0,0,0}, grass, false, false);
    grassFire(theImage, (Pixel){0,0,0}, grass, true, true);
  }
  
  resizeIm(theImage,theImage->cols/2,theImage->rows/2);
  resizeIm(theImage,theImage->cols/2,theImage->rows/2);
  resizeIm(theImage,theImage->cols/2,theImage->rows/2);
  
  /*resize to create the display image*/
  resizeIm(theImage,theImage->cols*2,theImage->rows*2);
  resizeIm(theImage,theImage->cols*2,theImage->rows*2);
  resizeIm(theImage,theImage->cols*2,theImage->rows*2);
  
  xim1->data=ConvertToX((Pixel *)theImage->image, xim1->data,Ysize, Xsize,bytesPerPixel);
  free(theImage->image);
  XPutImage(curDisplay, curWindow, curGC, xim1, 0, 0, 0, 0, Xsize, Ysize);
  
  return;
}

void closeWindow(long nFrames)
{

   XCloseDisplay(curDisplay);

   return;
}

/****************************************
 *        MAIN LOOP
 ****************************************/

int main(int argc, char * argv[])
{
   Bool   done = False,exam=false;
   XEvent curEvent;
   char c;
   long event_mask = 0xFFFFFFFF;
   long nBuf = 2;
   long start;
   bool go=true,loop=true,redo=false,setthresh=false;
   Image *grey,*tempimg;
   char filename[MAXBUF],target[MAXBUF],ldir[MAXBUF],buffer[MAXBUF];
   int count=0,i=0,temp,required=4,label[4];
   FILE *outfile;

   // default values
   grass=5;
   threshpad=70;
   thresh=(Pixel){96,84,74};
   while(loop){
     temp=getopt(argc,argv,"f:hs:t:l:");
     
     switch(temp){
     case 'f':
       strncpy(filename,optarg,sizeof(char)*MAXBUF);
       count++;
       break;
     case 's':
       strncpy(source,optarg,sizeof(char)*MAXBUF);
       count++;
       break;
     case 't':
       strncpy(target,optarg,sizeof(char)*MAXBUF);
       count++;
       break;
     case 'l':
       strncpy(ldir,optarg,sizeof(char)*MAXBUF);
       count++;
       break;
     case -1:
       if(count>=required){
	 loop=false;
	 break;
       }
     case ':': case 'h': case '?':
       printf("Usage: processTraining.bin -f <image list> -s <image dir> -t <target dir> -l <labeldir>\n");
       printf("\t <image list> should contain a list of all training files, without their .ppm suffix\n");
       exit(1);
       break;
       
     }
   }
   
   count=readfilenames(filename,list);
   /* initialize the window and video device */
   initWindow(nBuf);
   
   start = time(NULL);
   
   theImage=(Image *)malloc(sizeof(Image));
   theImage->color=true;

   while(!done) {
     if(i>=count){
       done=true;
       break;
     }
     // do any processing you need to do here
     // xim1->data points to the data (unsigned char *)
     if(go||redo){
       sprintf(buffer,"%s/%s.ppm",source,list[i]);
       theImage->image=readPPM(&theImage->rows,&theImage->cols,&temp,buffer);
       theImage->imagesize=theImage->rows*theImage->cols;

       /* create the training image*/
       gaussianIm(theImage);
       cropXYIm(theImage, 16, 12);

       tempimg=theImage;
       theImage=chromaticityIm(tempimg);
       free(tempimg->image);
       free(tempimg);
       threshByPaddedColor(theImage,thresh, threshpad);
       if(grass>0){
	 grassFire(theImage, (Pixel){0,0,0}, grass, false, false);
	 grassFire(theImage, (Pixel){0,0,0}, grass, true, true);
       }

       resizeIm(theImage,theImage->cols/2,theImage->rows/2);
       resizeIm(theImage,theImage->cols/2,theImage->rows/2);
       resizeIm(theImage,theImage->cols/2,theImage->rows/2);

       sprintf(buffer,"%s/%s.pgm",target,list[i]);
       grey=greyScaleIm(theImage);
       writePGM(grey->gimage,grey->rows,grey->cols,255,buffer);
       free(grey->gimage);
       free(grey);
       
       /*resize to create the display image*/
       resizeIm(theImage,theImage->cols*2,theImage->rows*2);
       resizeIm(theImage,theImage->cols*2,theImage->rows*2);
       resizeIm(theImage,theImage->cols*2,theImage->rows*2);
       
       xim1->data=ConvertToX((Pixel *)theImage->image, xim1->data,Ysize, Xsize,bytesPerPixel);
       free(theImage->image);
       XPutImage(curDisplay, curWindow, curGC, xim1, 0, 0, 0, 0, Xsize, Ysize);
       go=false;
       redo=false;
       label[0]=label[1]=label[2]=label[3]=0;
     }

     if(XCheckWindowEvent(curDisplay, curWindow, event_mask, &curEvent)) {
       
       switch (curEvent.type) {
	 
       case ButtonPress:
	 /*
	   access button by curEvent.xbutton.button (1,2,3)
	   access position by curEvent.xbutton.x and .y
	 */
	 if(exam){
	   sprintf(buffer,"%s/%s.ppm",source,list[i]);
	   theImage->image=readPPM(&theImage->rows,&theImage->cols,&temp,buffer);
	   theImage->imagesize=theImage->rows*theImage->cols;
	   gaussianIm(theImage);
	   cropXYIm(theImage, 16, 12);
	   tempimg=theImage;
	   theImage=chromaticityIm(tempimg);
	   free(tempimg->image);
	   free(tempimg);
	     
	   printf("X: %d, Y: %d :: r: %d g: %d\n",curEvent.xbutton.y,
		  curEvent.xbutton.y, 
		  theImage->image[curEvent.xbutton.y*theImage->cols
				  +curEvent.xbutton.x].r,
		  theImage->image[curEvent.xbutton.y*theImage->cols
				  +curEvent.xbutton.x].g);
	   free(theImage->image);
	   break;
	 }else if(setthresh){
	   if(curEvent.xbutton.button==1){
	     tempimg=theImage;
	     theImage=chromaticityIm(tempimg);
	     free(tempimg->image);
	     free(tempimg);
	     thresh=theImage->image[curEvent.xbutton.y*theImage->cols+curEvent.xbutton.x];
	     free(theImage->image);
	     printf("Threshold: %d %d %d\n",thresh.r,thresh.g,thresh.b);
	   }else{
	     printf("Threshold unchanged.\n");
	   }	   
	   setthresh=false;
	   redo=true;
	   break;
	 }
	 label[0]=1;
	 if(curEvent.xbutton.button == 1) {
	   label[1]=1;
	 }
	 else if(curEvent.xbutton.button == 2) {
	   label[1]=0;
	 }
	 else if(curEvent.xbutton.button == 3) {
	   label[1]=-1;
	 }
	 
	 label[2]=((double)curEvent.xbutton.x)/Xsize*36;
	 label[3]=((double)curEvent.xbutton.y)/Ysize*27;
	 
	 sprintf(buffer,"%s/%s.lab",ldir,list[i]);
	 outfile=fopen(buffer,"w");
	 fprintf(outfile,"%d %d %d %d\n",label[0],label[1],label[2],label[3]);
	 fclose(outfile);
	 i++;
	 go=true;
	 break;
	 
       case KeyPress:
	 /*
	   access string using XLookupString(*event,*char,numChars,NULL,NULL)
	 */
	 if(XLookupString((XKeyEvent *)&curEvent, &c, 1, NULL, NULL) == 1) {
	   
	   //	   printf("key code %02X\n", (int)c);
	   
	   switch(c) {
	   case 'b':
	     i--;
	     redo=true;
	     go=false;
	     break;
	   case 'f':
	     grass--;
	     printf("Grassfire level: %d\n",grass);
	     redo=true;
	     break;
	   case 'g':
	     grass++;
	     printf("Grassfire level: %d\n",grass);
	     redo=true;
	     break;
	   case 'q':
	     done = True;
	     break;
	   case 'n':
	     label[0]=0;
	     label[1]=0;
	     label[2]=0;
	     label[3]=0;
	     sprintf(buffer,"%s/%s.lab",ldir,list[i]);
	     outfile=fopen(buffer,"w");
	     fprintf(outfile,"%d %d %d %d\n",label[0],label[1],label[2],label[3]);
	     fclose(outfile);

	     i++;
	     go=true;
	     break;
	   case 't':
	     if(exam)
	       break;
	     setthresh=true;
	     printf("Click left mouse button to set the threshold value...\n");
	     sprintf(buffer,"%s/%s.ppm",source,list[i]);
	     theImage->image=readPPM(&theImage->rows,&theImage->cols,&temp,buffer);
	     theImage->imagesize=theImage->rows*theImage->cols;
	   

	     gaussianIm(theImage);
	     cropXYIm(theImage, 16, 12);

	     /*	     tempimg=theImage;
	     theImage=chromaticityIm(tempimg);
	     free(tempimg->image);
	     free(tempimg);
	     */
	     resizeIm(theImage,theImage->cols/2,theImage->rows/2);
	     resizeIm(theImage,theImage->cols/2,theImage->rows/2);
	     resizeIm(theImage,theImage->cols/2,theImage->rows/2);
	     resizeIm(theImage,theImage->cols*2,theImage->rows*2);
	     resizeIm(theImage,theImage->cols*2,theImage->rows*2);
	     resizeIm(theImage,theImage->cols*2,theImage->rows*2);
	     
	     xim1->data=ConvertToX((Pixel *)theImage->image, xim1->data,Ysize, Xsize,bytesPerPixel);
	     
	     XPutImage(curDisplay, curWindow, curGC, xim1, 0, 0, 0, 0, Xsize, Ysize);

	     break;
	   case '+':
	     threshpad++;
	     printf("Threshold padding: %d\n",threshpad);
	     redo=true;
	     break;
	   case '-':
	     threshpad--;
	     printf("Threshold padding: %d\n",threshpad);
	     redo=true;
	     break;
	   case 'x':
	     if(exam){
	       printf("Examination mode turned off.\n");
	       exam=false;
	     }else{
	       exam=true;
	       printf("Examination mode turned on.\n");
	     }
	   default:
	     break;
	   }
	 }
	 break;
	 
       case Expose:
	 refreshWindow(i);
	 break;
	 
       case ResizeRequest:
	 XResizeWindow(curDisplay, curWindow, Xsize, Ysize);
	 printf("NO RESIZE FOR YOU!\n");
	 break;
	 
       default:
	 break;
       }
     }
     
   }
   
   closeWindow(nBuf);
   
   
   return(0);
}

