/* Canny Edge Detector from Computer Vision by Shapiro and Stockman p 300 */
#include <visenv.h>

Image Canny(Image *image, int Tlow, int Thigh) {

  unsigned char *pgm;
  unsigned char *smooth;
  unsigned char *edge;
  double *op[2];
  Image theImage;
  int i;

  op[0] = (double *)calloc(image->imagesize,sizeof(double));
  op[1] = (double *)calloc(image->imagesize,sizeof(double));

  printf("smoothing the image\n");
  pgm = ppmToPGM(image);
  smooth = applyGaussian(pgm, image->rows, image->cols);
  free(pgm);

  printf("applying Sobel Operators\n");
  SobelOp(smooth, image->rows, image->cols, op[1], op[2]);

  printf("suppressing nonmaxima\n");
  SuppressNonmaxima(op[0], op[1], image->rows, image->cols);

  printf("detecting edges\n");
  edge = EdgeDetect(op[0], Tlow, Thigh, image->rows, image->cols);

  printf("writing the image\n");
  theImage.rows = image->rows;
  theImage.cols = image->cols;
  theImage.imagesize = image->imagesize;
  theImage.image = (Pixel *)malloc(sizeof(Pixel)*theImage.imagesize);

  for(i=0; i<theImage.imagesize; i++)
    theImage.image[i] = (Pixel){edge[i],edge[i],edge[i]};

  free(op[0]);
  free(op[1]);

  return theImage;
}

unsigned char *ppmToPGM(Image *image) {

  unsigned char *theImage = (unsigned char *)malloc(sizeof(unsigned char)*image->imagesize);
  int i;

  for(i=0; i<image->imagesize; i++) {
    theImage[i] = (unsigned char)(((double)image->image[i].r+(double)image->image[i].g
				  +(double)image->image[i].b)/3.0);
  }

  return theImage;
}

//Gaussian smoothing by 5x5 filter
unsigned char *applyGaussian(unsigned char *image, int rows, int cols) {

  unsigned char *theImage = (unsigned char *)malloc(sizeof(unsigned char)*rows*cols);
  int i, j, x, y;
  double gaussian[5][5] = {{1,2,4,2,1},
			   {2,4,6,4,2},
			   {4,6,10,6,4},
			   {2,4,6,4,2},
			   {1,2,4,2,1}};
  double sumG, sumT;

  for(j=0; j<rows; j++) {
    for(i=0; i<cols; i++) {
      theImage[j*cols+i]=0;
      sumG = sumT = 0.0;
      for(y=0; y<=5; y++) {
	if(j+y-2>=0 && j+y-2<rows) {
	  for(x=0; x<=5; x++) {
	    if(i+x-2>=0 && i+x-2<cols) {
	      sumG += ((double)image[(j+y-2)*cols+(i+x-2)]*gaussian[x][y]);
	      sumT += gaussian[x][y];
	    }
	  }
	}
      }
      theImage[j*cols+i] = (unsigned char)(sumG/sumT);
    }
  }

  return theImage;
}
// This function suppresses non-maxima
void SuppressNonmaxima(double *mag, double *dir, int rows, int cols) {
  
  int pDel[4][2] = {{1,0},{1,1},{0,1},{-1,1}};
  int nDel[4][2] = {{-1,0},{-1,-1},{0,-1},{1,-1}};
  int direction;
  int i, j;

  for(i=1; i<rows-1; i++) {
    for(j=1; j<cols-1; j++) {
      direction = dir[i*rows+j];
      if(mag[i*rows+j] <= mag[(int)(i+pDel[direction][1])*rows+ 
			     j+pDel[direction][0]])
	mag[i*rows+j] = 0.0;
      if(mag[i*rows+j] <= mag[(int)(i+nDel[direction][1])*rows+
			     j+nDel[direction][0]])
	mag[i*rows+j] = 0.0;
    }
  }
}

unsigned char *EdgeDetect(double *mag, int low, int high, int rows, int cols) {

  unsigned char *theImage = (unsigned char *)calloc(rows*cols,sizeof(unsigned char));
  bool *imageMap = (bool *)malloc(sizeof(bool)*rows*cols);
  int i, j;

  for(i=0; i<rows*cols; i++) {
    imageMap[i] = false;
  }

  for(i=1; i<rows-1; i++) {
    for(j=1; j<cols-1; j++) {
      if(mag[i*cols+j] >= high)
	FollowEdge(i,j,rows,cols,mag,low,high,imageMap);
    }
  }

  //set the pixel white if an edge pixel, black otherwise
  for(i=0; i<rows*cols; i++) {
    theImage[i] = imageMap[i]==true ? 255 : 0;
  }

  return theImage;
}

void FollowEdge(int row, int col, int rows, int cols, double *mag, int low, 
		int high, bool *image) {
  int u, v, i;
  bool flag=true;
  
  printf("row: %d\tcol: %d\n",row, col);
  u=row; v=col;

  while(flag) {
    flag=false;
    image[u*cols+v]=true;
    row=u; col=v;
    //check 8-neighbors (top left to bottom right)
    for(i=0; i<8; i++) {
      u = j<3 ? row-1 : j<5 ? row : row+1;
      v = j<3 ? col+j-1 : j==3 ? col-1 : j==4 ? col+1 : col+j-6;
      //if within the image
      if(u>0 && u<rows && v>0 && v<cols) {
	//if greater than lower threshold
	if(mag[u*cols+v]>low) {
	  flag = true;
	  break;
	}
      }
    }//end for loop
  }

}
