/**********************************
 * Functions to convert to different
 * color spaces. 
 * HSV functions from: http://www.cs.rit.edu/~ncs/color/t_convert.html
 * Transforms also from aforementioned site.
 **********************************/

/*********************
 * HSV functions
 *********************/

#include <visenv.h>

// r,g,b values are from 0 to 1
// h = [0,360], s = [0,1], v = [0,1]
//              if s == 0, then h = -1 (undefined)
void RGBtoHSV( float r, float g, float b, float *h, float *s, float *v )
{
        float min, max, delta;
	min=r;
	if(g<min)
	  min=g;
	if(b<min)
	  min=b;

	max=r;
	if(g<max)
	  max=g;
	if(b<max)
	  max=b;

        *v = max;                               // v
        delta = max - min;
        if( max != 0 )
                *s = delta / max;               // s
        else {
                // r = g = b = 0                // s = 0, v is undefined
                *s = 0;
                *h = -1;
                return;
        }
        if( r == max )
                *h = ( g - b ) / delta;         // between yellow & magenta
        else if( g == max )
                *h = 2 + ( b - r ) / delta;     // between cyan & yellow
        else
                *h = 4 + ( r - g ) / delta;     // between magenta & cyan
        *h *= 60;                               // degrees
        if( *h < 0 )
                *h += 360;
}

/*****************************
 * Following functions written
 * by fritz heckel. YIQ and CIE
 *****************************/

Pixel colorTransform(Pixel color,double transform[9]){

  double normalized[3],transformed[3];
  Pixel retval;

  int i;

  normalized[0]=(double)color.r/255.0;
  normalized[1]=(double)color.g/255.0;
  normalized[2]=(double)color.b/255.0;

  for(i=0;i<3;i++){
    transformed[i]=
      transform[i*3+0]*normalized[0]+
      transform[i*0+1]*normalized[1]+
      transform[i*0+2]*normalized[2];
    
  }

  retval.r=(unsigned char)(transformed[0]*255);
  retval.g=(unsigned char)(transformed[1]*255);
  retval.b=(unsigned char)(transformed[2]*255);

  return retval;

}

Pixel toYIQ(Pixel color){

  double transform[9]={0.299,  0.587,  0.114,
		       0.596, -0.275, -0.321,
		       0.212, -0.523,  0.311};

  return colorTransform(color,transform);

}

Pixel toXYZ(Pixel color){

  double transform[9]={0.412453, 0.357580, 0.180423,
		       0.212671, 0.715160, 0.072169, 
		       0.019334, 0.119193, 0.950227};

  return colorTransform(color,transform);

}

Image *HSVImage(Image *theImage){

  int i;
  float temp[3];

  Image *newImage;
  newImage=(Image *)malloc(sizeof(Image));
  newImage->color=true;
  newImage->chroma=false;
  newImage->rows=theImage->rows;
  newImage->cols=theImage->cols;
  newImage->imagesize=theImage->imagesize;
  newImage->image=(Pixel *)malloc(sizeof(Pixel)*newImage->imagesize);

  if(!newImage->color||newImage->chroma){
    printf("Fatal: Greyscale or chromaticity image passed to HSVImage!\n");
    exit(-1);
  }

  for(i=0;i<newImage->imagesize;i++){
    RGBtoHSV(theImage->image[i].r/255.0,
	     theImage->image[i].g/255.0,
	     theImage->image[i].b/255.0,
	     &temp[1],&temp[2],&temp[3]);
    newImage->image[i]=(Pixel){(unsigned char)((temp[1]/360.0)*255),
			       (unsigned char)(temp[2]*255),
			       (unsigned char)(temp[3]*255)};
  }
  return newImage;
}

Image *YIQImage(Image *theImage){

  int i;

  Image *newImage;
  newImage=(Image *)malloc(sizeof(Image));
  newImage->color=true;
  newImage->chroma=false;
  newImage->rows=theImage->rows;
  newImage->cols=theImage->cols;
  newImage->imagesize=theImage->imagesize;
  newImage->image=(Pixel *)malloc(sizeof(Pixel)*newImage->imagesize);

  
  if(!newImage->color||newImage->chroma){
    printf("Fatal: Greyscale or chromaticity image passed to YIQImage!\n");
    exit(-1);
  }

  for(i=0;i<newImage->imagesize;i++){
    newImage->image[i]=toYIQ(theImage->image[i]);
  }

  return newImage;

}

Image *XYZImage(Image *theImage){

  int i;

  Image *newImage;
  newImage=(Image *)malloc(sizeof(Image));
  newImage->color=true;
  newImage->chroma=false;
  newImage->rows=theImage->rows;
  newImage->cols=theImage->cols;
  newImage->imagesize=theImage->imagesize;
  newImage->image=(Pixel *)malloc(sizeof(Pixel)*newImage->imagesize);
  
  if(!newImage->color||newImage->chroma){
    printf("Fatal: Greyscale or chromaticity image passed to XYZImage!\n");
    exit(-1);
  }

  for(i=0;i<newImage->imagesize;i++){
    newImage->image[i]=toXYZ(newImage->image[i]);
  }

  return newImage;

}
