#include <visenv.h>

int *intensityHistogram(Image *theImage) {

  int i, *histogram,temp;

  histogram=(int *)calloc(256,sizeof(int));
  
  for(i=0;i<theImage->imagesize;i++){
    temp=(int)(((int)(theImage->image[i].r)+
		(int)(theImage->image[i].g)+
		(int)(theImage->image[i].b))/3.0);
    if(temp<0||temp>255){
      printf("The stack corruption happens in intensityHistogram()\n");
    }else{
      histogram[temp]=histogram[temp]+1;      
    }
  }
  return histogram;
}

int *color2DHistogram(Image *theImage) {

  int i, *histogram;

  if((!theImage->color)&&(!theImage->chroma)){
    printf("Passed a grey image to color2DHistogram. Fatal.\n");
    exit(-1);
  }

  histogram=(int *)calloc(256*256,sizeof(int));

  for(i=0;i<theImage->imagesize;i++) {
    histogram[(int)theImage->image[i].r*256+theImage->image[i].g]++;  
  }

  return histogram;
}

int *color2DsubHistogram(Image *theImage,int top, int left, int bot, int right) {

  int i, j,*histogram;

  if((!theImage->color)&&(!theImage->chroma)){
    printf("Passed a grey image to color2DsubHistogram. Fatal.\n");
    exit(-1);
  }

  histogram=(int *)calloc(256*256,sizeof(int));

  for(i=top;i<=bot;i++) {
    for(j=left;j<=right;j++){
      histogram[(int)theImage->image[i*theImage->cols+j].r*256+theImage->image[i*theImage->cols+j].g]++;  
    }
  }

  return histogram;
}

int *color3DHistogram(Image *theImage) {

  int *histogram;
  long i;

  if((!theImage->color)&&(!theImage->chroma)){
    printf("Passed a grey image to color3DHistogram. Fatal.\n");
    exit(-1);
  }

  histogram=(int *)calloc(256*256*256,sizeof(int));

  for(i=0;i<theImage->imagesize;i++) {
    histogram[(int)theImage->image[i].r*256*256+theImage->image[i].g*256+theImage->image[i].b]++;  
  }

  return histogram;
}

int *color3DsubHistogram(Image *theImage,int top, int left, int bot, int right) {

  int i, j,*histogram;
  
  if((!theImage->color)&&(!theImage->chroma)){
    printf("Passed a grey image to color3DHistogram. Fatal.\n");
    exit(-1);
  }

  histogram=(int *)calloc(256*256*256,sizeof(int));

  for(i=top;i<=bot;i++) {
    for(j=left;j<=right;j++){
      histogram[(int)theImage->image[i*theImage->cols+j].r*256*256+theImage->image[i*theImage->cols+j].g*256+theImage->image[i*theImage->cols+j].b]++;  
    }
  }

  return histogram;
}

int *blurHistogram(int *histogram, int size) {
  
  int *outputHistogram;
  long i;
  outputHistogram=(int *)calloc(size,sizeof(int));

  for(i=0;i<size;i++){
    if(i==0){
      outputHistogram[i]=(histogram[i]*2+histogram[i+1])/3.0;
    }else if(i==size){
      outputHistogram[i]=(histogram[i]*2+histogram[i-1])/3.0;
    }else{
      outputHistogram[i]=(histogram[i-1]+histogram[i]*2+
			  histogram[i-1])/4.0;      
    }
  }

  return outputHistogram;
}

int *maskHistogram(Image *theImage, Image *mask) {

  int i, *histogram;
  histogram=(int *)calloc(256,sizeof(int));
  
  for(i=0;i<theImage->imagesize;i++){
    if(mask->image[i].r != 0 &&
       mask->image[i].g != 0 &&
       mask->image[i].b != 0) {
      histogram[(int)(((int)(theImage->image[i].r) + 
		 (int)(theImage->image[i].g) + 
		 (int)(theImage->image[i].b))/3.0)]+=1;        
    }
  }

  return histogram;
}

int findThreshold(int *histogram, int size, int guess, double factor) {
  
  double bgmean, fgmean;
  int i, thresh, bgcount, fgcount;
  
  thresh = guess;  
  
  while(1) {
    bgmean = fgmean = bgcount = fgcount = 0;

    for(i=0; i<thresh; i++) {
      bgmean = bgmean + histogram[i] * i;
      bgcount = bgcount + histogram[i];
    }
    bgmean /= (double)bgcount;

    for(i=size-1; i>thresh; i--) {
      fgmean = fgmean + histogram[i] * i;
      fgcount = fgcount + histogram[i];
    }
    fgmean /= (double)fgcount;

    if(thresh == (int)((bgmean + fgmean)/2.0))
      break;
    else
      thresh = (int)((bgmean + fgmean)/2.0);
    if(thresh<0||thresh>255){
      thresh=180;
      break;
    }
    printf("thresh %d\n", thresh);
  }
  
  
  return thresh;
}

int *minMaxThresh(int *histogram, int threshold, int size) {
  
  int i, *minMax = (int *)malloc(sizeof(int)*2);

  for(i=0; i<size; i++) {
    if(histogram[i] > 0) {
      minMax[0] = i;
      break;
    }
  }
  minMax[0] += (threshold - minMax[0]) * 0.5;

  for(i=size-1; i>=0; i--) {
    if(histogram[i] > 0) {
      minMax[1] = i;
      break;
    }
  }
  minMax[1] -= (minMax[1] - threshold) * 0.5;

  return minMax;
}

double histIntersect(double *hist1, double *hist2, long size){

  long i;
  double sum=0;

  for(i=0;i<size;i++){
    sum+=hist1[i]<hist2[i]?hist1[i]:hist2[i];
  }

  return sum;

}

double histLNorm(double *hist1, double *hist2, long size, int power){

  long i;
  double norm=0,temp;

  for(i=0;i<size;i++){
    temp=hist1[i]-hist2[i];
    if(temp<0)
      temp*=-1;

    norm+=pow(temp,(double)power)*100;
  }

  return norm;
  
}

double * normalizeHistogram(int *histogram, long bins){

  long sum=0;
  long i;
  double *normalized=malloc(sizeof(double)*bins);

  for(i=0;i<bins;i++){
    sum+=histogram[i];
  }

  for(i=0;i<bins;i++){
    normalized[i]=((double)histogram[i])/((double)sum);
  }

  return normalized;

}

int *resizeHistogram(int *histogram, long size) {

  int i, j;
  int *temp=calloc(size,sizeof(int));
  int *resized=(int *)malloc(sizeof(int)*100);
  int gaussian[5]={1,2,4,2,1};
  float delta;

  //Gaussian blurring
  for(i=2; i<size-2; i++) {
    for(j=0;j<5;j++) {
      temp[i]+=gaussian[j]*histogram[i+j-2];
    }
    temp[i]/=10;
  }

  //Resizing histogram
  delta = (float)size/100.0;
  for(i=0; i<100; i++) {
    resized[i]=temp[(int)(i*delta)];
  }

  free(temp);
  return resized;
}
