#include <visenv.h>
#include <string.h>

/* writes count vectors of size bins to a file. note that things are slightly
   backwards. */
void writeVectorsToFile(double **basis, int count, int size,char *filename){

  int i,j;

  FILE *outfile=fopen(filename, "w");

  fprintf(outfile,"%d %d\n",count,size);

  for(i=0;i<count;i++){
    for(j=0;j<size;j++){
      fprintf(outfile,"%f ",basis[j][i]);
    }
    fprintf(outfile,"\n");
  }

  fclose(outfile);
}

/* reads vectors from file. vectors in the same backwards format from
   above */
double **readVectorsFromFile(int *count, int *size, char *filename){

  int i,j;
  double **retval,*imageMatrix;
  
  FILE *infile=fopen(filename, "r");

  fscanf(infile,"%d %d\n",count,size);

  // allocate space for retval
  retval=(double **)malloc(sizeof(double *)**size);
  imageMatrix = (double *)malloc(sizeof(double)**size**count);
  for(j=0;j<*size;j++)
    retval[j]=&(imageMatrix[j**count]);

  for(i=0;i<*count;i++){
    for(j=0;j<*size;j++){
      fscanf(infile,"%lf ",&retval[j][i]);
    }
    fscanf(infile,"\n");
  }

  return retval;
}

int readfilenames(char *listfile, char filenames[MAXIMS][MAXBUF]){

  FILE * infile=fopen(listfile,"r");
  char buffer[256],*temp;

  int i=0;

  if(!infile){
    printf("Fatal: Unable to open listfile %s\n",listfile);
    exit(-1);
  }

  while(fgets(buffer,sizeof(char)*MAXBUF,infile)){
    temp=strstr(buffer,"\n");

    if(temp)
      *temp='\0';

    strncpy(filenames[i],buffer,sizeof(char)*MAXBUF);
    i++;
  }

  return i;
}

void writeFaceModelsToFile(Face *faceModels, int count, char *filename){

  int i,j;

  FILE *outfile=fopen(filename, "w");

  if(!outfile){
    printf("Fatal: Cannot open facemodel file %s for writing.\n",filename);
    exit(-1);
  }

  fprintf(outfile,"%d %d\n",count,faceModels[0].size);

  for(i=0;i<count;i++){

    fprintf(outfile,"%s ",faceModels[i].label);
    for(j=0;j<faceModels[i].size;j++){
      fprintf(outfile,"%f ",faceModels[i].model[j]);
    }
    fprintf(outfile,"\n");
  }

  fclose(outfile);

}

Face *readFaceModelsFromFile(int *count, char *filename){

  int i,j,size;
  Face *retval;
  FILE *infile=fopen(filename, "r");

  if(!infile){
    printf("Fatal: Cannot open facemodel file %s for reading.\n",filename);
    exit(-1);
  }

  fscanf(infile,"%d %d\n",count,&size);

  retval=(Face *)malloc(sizeof(Face)**count);

  for(i=0;i<*count;i++){

    fscanf(infile,"%s ",retval[i].label);
    retval[i].model=(double *)malloc(sizeof(double)*size);
    retval[i].size=size;
    for(j=0;j<size;j++){
      fscanf(infile,"%lf ",&retval[i].model[j]);
    }
    fscanf(infile,"\n");
  }

  fclose(infile);

  return retval;

}

LabeledFile *readLabeledFileList(char *listfile, int *count){

  FILE * infile=fopen(listfile,"r");
  LabeledFile *retval;

  int i=0;

  if(!infile){
    printf("Fatal: Unable to open listfile %s\n",listfile);
    exit(-1);
  }

  fscanf(infile,"%d\n",count);

  retval=(LabeledFile *)malloc(sizeof(LabeledFile)**count);
  
  for(i=0;i<*count;i++){

    fscanf(infile,"%s %s\n",retval[i].path,retval[i].label);

  }
  
  fclose(infile);

  return retval;

}
