#include <visenv.h>


/************************
 * Image Conversion Functions
 *************************/
Image *pgmToPPM(unsigned char *image, int rows, int cols){

  int i;

  Image *theImage=(Image *)malloc(sizeof(Image));

  theImage->rows=rows;
  theImage->cols=cols;
  theImage->imagesize=rows*cols;

  theImage->image=(Pixel *)malloc(sizeof(Pixel)*rows*cols);
  
  for(i=0;i<theImage->imagesize;i++){
    theImage->image[i].r=image[i];
    theImage->image[i].g=image[i];
    theImage->image[i].b=image[i];
  }

  return theImage;

}

void genGrayFromColor(Image *image){

  int i;

  image->gimage=(unsigned char *)malloc(sizeof(unsigned char)*image->imagesize);
  if(!image->gimage){
    fprintf(stderr,"Fatal error: Out of memory in genGrayFromColor()\n");
    exit(-1);
  }

  for(i=0;i<image->imagesize;i++){
    image->gimage[i]=(image->image[i].r+image->image[i].g+image->image[i].b)/3;
  }
}

/*********************
 * Sobel Operator		      
 *********************/

// Calculate magnitude and edge direction using Sobel Operators
// http://www.pages.drexel.edu/~weg22/can_tut.html
void SobelOp(unsigned char *image, int rows, int cols, double *mag, double *dir) {

  double gx, gy, orient;
  double mgx[3][3] = {{-1,0,1},{-2,0,2},{-1,0,1}};
  double mgy[3][3] = {{1,2,1},{0,0,0},{-1,-2,-1}};
  int i, j, a, b;

  for(i=1; i<rows-1; i++) {
    for(j=1; j<cols-1; j++) {
      //calculate the magnitude of the gradient
      gx = gy = orient = 0.0;
      for(a=0; a<3; a++) {
	for(b=0; b<3; b++) {
	  gx = (double)image[(i-1+a)*cols+(j-1+a)]*mgx[a][b];
	  gy = (double)image[(i-1+a)*cols+(j-1+a)]*mgy[a][b];
	}
      mag[i*cols+j] = abs(gx) + abs(gy);
      if(mag[i*cols+j] > 255)
	mag[i*cols+j] = 255;

      //calculate the direction of the gradient
      orient = atan2(gy,gx)*PI_CONVERT;
      dir[i*rows+j] = (orient < 22.5 ? 0 : orient < 67.5 ? 1 : 
		       orient < 112.5 ? 2 : orient < 157.5 ? 3 : 0); 
      }
    }
  }
}

/********************
 * Masking Functions
 ********************/

Image *maskBackground(Image *image, int guess, int factor, bool inverse){

  Image *output, *chrom;
  int *ihistogram, *temp, i, threshold;
  double sum;
  bool tflag;
  Pixel tColor, iColor;

  output=newImage(image->rows,image->cols,false,true);

  // build the intensity histogram and blur it
  ihistogram = intensityHistogram(image);
  temp=ihistogram;
  ihistogram=blurHistogram(temp,255);
  free(temp);

  threshold = findThreshold(ihistogram,255,guess,factor);
  printf("threshold %d\n",threshold);

  //create chromaticity image
  chrom = chromaticityIm(image);

  tflag=true;
#ifdef DEBUG
  if(tflag) printf("tflag: true\n");
  printf("DEBUG: Writing chromaticity image\n");
  writePPM(chrom->image,chrom->rows,chrom->cols,255,"chrom.ppm");
#endif
  
  for(i=0; i<output->imagesize; i++) {
    sum = ((double)image->image[i].r+(double)image->image[i].g+(double)image->image[i].b); 
    if(sum/3.0 > threshold) {
      output->image[i]= (Pixel){0,0,0};
      if(tflag && sum/3.0<threshold+5) {
	tflag=false;
	//used to compare chromaticity
	tColor.r = ((double)image->image[i].r/(double)sum)*255;
	tColor.g = ((double)image->image[i].g/(double)sum)*255;
	tColor.b = ((double)image->image[i].b/(double)sum)*255;
	iColor = image->image[i]; //used to compare intensity
#ifdef DEBUG
	printf("tColor.r %u\ttColor.g %u\ttColr.b %u\n",tColor.r,tColor.g,tColor.b);
	printf("iColor.r %u\tiColor.g %u\tiColr.b %u\n",iColor.r,iColor.g,iColor.b);
#endif
      }
    }
    else {
      //ignore this for now
      // if the changes in chromaticity is gradual 
      /* consider current pixel as background
      if(chrom->image[i].r>tColor.r-RANGE && chrom->image[i].r<tColor.r+RANGE &&
	 chrom->image[i].g>tColor.g-RANGE && chrom->image[i].g<tColor.g+RANGE &&
	 chrom->image[i].b>tColor.b-RANGE && chrom->image[i].b<tColor.b+RANGE &&
	 image->image[i].r>iColor.r-30 && image->image[i].g>iColor.g-30
	 && image->image[i].b>iColor.b-30) {
	output->image[i] = (Pixel){0,0,0};
      } 
      else { 
      */
	output->image[i] = (Pixel){255,255,255};	  
	//}
    }
  }

  //free images
  free(ihistogram);
  free(chrom->image);
  free(chrom);

  return output;
}

/******************
 * Miscellaneous
 ********************/

bool checkPixel(Image *image,int index,Pixel value){

  if(image->color||image->chroma){
    if(index>=image->imagesize){
      printf("Overflow.\n");
      exit(-1);
    }
    if(image->image[index].r!=value.r||
       image->image[index].g!=value.g||
       image->image[index].b!=value.b){
      return false;
    }
  }else{
    if(image->gimage[index]!=value.r){
      return false;
    }
  }

  return true;
}

void setPixel(Image *image,int index,Pixel value){

  if(index>=image->imagesize){
    fprintf(stderr,"Index out of bounds in setPixel\n");
    return;
  }

  if(image->color||image->chroma){
    image->image[index].r=value.r;
    image->image[index].g=value.g;
    image->image[index].b=value.b;
  }else{
    image->gimage[index]=value.r;
  }

}

bool is4Boundary(Image *image, int x, int y, Pixel bgVal){

  if((y>0&&checkPixel(image,(y-1)*image->cols+x,bgVal))||
     (y<image->rows-1&&checkPixel(image,(y+1)*image->cols+x,bgVal))||
     (x<image->cols-1&&checkPixel(image,y*image->cols+(x+1),bgVal))||
     (x>0&&checkPixel(image,y*image->cols+(x-1),bgVal))){
    return true;
  }

  return false;
}

bool is8Boundary(Image *image, int x, int y, Pixel bgVal){

  int i, j;
  
  for(i=y-1;i<=y+1;i++){
    for(j=x-1;j<=x+1;j++){
      if(i>=0&&j>=0&&i<image->rows&&j<image->cols&&
      	 checkPixel(image,(i*image->cols)+j,bgVal)){
	return true;
      }
    }
  }

  return false;
}

short upLeftMin8(short *tmp, int y, int x,int rows,int cols){

  short vals[4]={-1,-1,-1,-1},min;

  if(x-1>=0){
    vals[0]=tmp[y*cols+x-1];
    if(y-1>=0){
      vals[1]=tmp[(y-1)*cols+x-1];
    }
  }

  if(y-1>=0){
    vals[2]=tmp[(y-1)*cols+x];
    if(x+1<cols){
      vals[3]=tmp[(y-1)*cols+x+1];
    }
  }
  
  min=vals[0];
  if(vals[1]<min)
    min=vals[1];
  if(vals[2]<min)
    min=vals[2];
  if(vals[3]<min)
    min=vals[3];
		
  return min;
}

short downRightMin8(short *tmp, int y, int x, int rows, int cols){

  short vals[4]={-1,-1,-1,-1},min;

  if(x+1<cols){
    vals[0]=tmp[y*cols+x+1];
    if(y+1<rows){
      vals[1]=tmp[(y+1)*cols+x+1];
    }
  }

  if(y+1<rows){
    vals[2]=tmp[(y+1)*cols+x];
    if(x+1<cols){
      vals[3]=tmp[(y+1)*cols+x-1];
    }
  }
  
  min=vals[0];
  if(vals[1]<min)
    min=vals[1];
  if(vals[2]<min)
    min=vals[2];
  if(vals[3]<min)
    min=vals[3];
		
  return min;
}

short upLeftMin4(short *tmp, int y, int x,int rows,int cols){

  short vals[2]={-1,-1};

  if(x-1>=0){
    vals[0]=tmp[y*cols+x-1];
  }

  if(y-1>=0){
    vals[1]=tmp[(y-1)*cols+x];
  }

  if(vals[0]<=vals[1]){
    return vals[0];
  }else{
    return vals[1];
  }

}

short downRightMin4(short *tmp, int y, int x,int rows,int cols){

  short vals[2]={-1,-1};

  if(x+1<cols){
    vals[0]=tmp[y*cols+x+1];
  }

  if(y+1<rows){
    vals[1]=tmp[(y+1)*cols+x];
  }
		
  return (vals[0]<=vals[1]?vals[0]:vals[1]);
}

Image *newImage(int rows,int cols, bool chroma, bool color){

  Image * newimg;

  newimg=(Image *)malloc(sizeof(Image));

  if(!newimg){
    fprintf(stderr,"Fatal error: out of memory in newImage()\n");
    exit(-1);
  }

  newimg->rows=rows;
  newimg->cols=cols;
  newimg->imagesize=rows*cols;
  newimg->chroma=chroma;
  newimg->color=color;

  if(color||chroma){
    newimg->image=(Pixel *)malloc(sizeof(Pixel)*cols*rows);
    if(!newimg->image){
      fprintf(stderr,"Fatal error: out of memory in newImage()\n");
      exit(-1);
    }
  }else{
    newimg->gimage=(unsigned char *)calloc(rows*cols, sizeof(unsigned char *));
    if(!newimg->gimage){
      fprintf(stderr,"Fatal error: out of memory in newImage()\n");
      exit(-1);
    }
  }

  return newimg;

}

void trimImage(Image *masked,Pixel bgVal,int border){

  int i, j;

  for(i=0;i<masked->rows;i++){
    if(i<border || i>=masked->rows-border) {
      for(j=0;j<masked->cols;j++){      
	setPixel(masked,i*masked->cols+j,bgVal);
      }
    }
    else {
      for(j=0;j<masked->cols;j++) {
	if(j<border || j>=masked->cols-border) {
	  setPixel(masked,i*masked->cols+j,bgVal);
	}
      }
    }
  }
  
}
